@@ -1,5 +1,5 @@
 /*
- * fsck.c - a file system consistency checker for Linux.
+ * fsck.minix.c - a file system consistency checker for Linux.
  *
  * (C) 1991, 1992 Linus Torvalds. This file may be redistributed
  * as per the GNU copyleft.
@@ -100,10 +100,12 @@
 #include <termios.h>
 #include <mntent.h>
 #include <sys/stat.h>
+#include <sys/param.h>
 #include <signal.h>
 
 #include "minix.h"
 #include "nls.h"
+#include "pathnames.h"
 
 #ifndef __linux__
 #define volatile
@@ -168,10 +170,8 @@ static unsigned char * zone_count = NULL;
 static void recursive_check(unsigned int ino);
 static void recursive_check2(unsigned int ino);
 
-#include "bitops.h"
-
-#define inode_in_use(x) (bit(inode_map,(x)))
-#define zone_in_use(x) (bit(zone_map,(x)-FIRSTZONE+1))
+#define inode_in_use(x) (isset(inode_map,(x)))
+#define zone_in_use(x) (isset(zone_map,(x)-FIRSTZONE+1))
 
 #define mark_inode(x) (setbit(inode_map,(x)),changed=1)
 #define unmark_inode(x) (clrbit(inode_map,(x)),changed=1)
@@ -179,7 +179,6 @@ static void recursive_check2(unsigned int ino);
 #define mark_zone(x) (setbit(zone_map,(x)-FIRSTZONE+1),changed=1)
 #define unmark_zone(x) (clrbit(zone_map,(x)-FIRSTZONE+1),changed=1)
 
-
 static void
 reset(void) {
 	if (termios_set)
@@ -304,7 +303,7 @@ check_mount(void) {
 	int cont;
 	int fd;
 
-	if ((f = setmntent (MOUNTED, "r")) == NULL)
+	if ((f = setmntent (_PATH_MOUNTED, "r")) == NULL)
 		return;
 	while ((mnt = getmntent (f)) != NULL)
 		if (strcmp (device_name, mnt->mnt_fsname) == 0)
@@ -318,7 +317,7 @@ check_mount(void) {
 	 * probably not correct; so we won't issue a warning based on
 	 * it.
 	 */
-	fd = open(MOUNTED, O_RDWR);
+	fd = open(_PATH_MOUNTED, O_RDWR);
 	if (fd < 0 && errno == EROFS)
 		return;
 	else
